/*
 *
 *
 */
#include "common.h"
#include "usb.h"
#include "usbhw.h"

/*
 * internal buffer size for endpoint 0
 *   must be greater than max_packet_size described in device desctiptor
 */
#define EP0_BUFF_SIZE    64

/*
 * Max Number of Interfaces
 */
#define USB_N_IF          4

/*
 * select default power setting
 *   0: Bus-powered
 *   1: Self-powered
 */
#define USB_POWER         0

/*
 * make state machine for CONTROL-PIPE
 */
#define CPS_IDLE     0		/* only SETUP will be accepted */
#define CPS_DATAIN   1		/* expect IN from host */
#define CPS_DATAOUT  2		/* expect OUT from host */
#define CPS_STATOUT  3		/* expect OUT from host */
#define CPS_STATIN   4		/* expect IN from host */
#if 0
[SUDS_IDLE]
 | SETUP(OUT)
 |     WITH DATA
				[SUDS_DATAOUT]
 |         IN->stall
 |         OUT
 |             ->stall if buffer over
 |             ->recv data until specified length--> ep0_data_ready=TRUE
						[SUDS_STATIN]
 |                 IN->send status
 |                 OUT->stall
 |     WITHOUT DATA
				[SUDS_STATIN] <========
 |         IN->send status
 |         OUT->stall
 | SETUP(IN)
 |     WITHOUT DATA
				[ucore_STALL_EP0()]
 |         IN->stall
 |         OUT->stall
 |     WITH DATA
				[SUDS_DATAIN] <=========
 |         OUT->stall
 |         IN->send data (zero-length-packet if needed)
										(datasize<requestsize) &&
										(datasize % packetsize == 0)
					[SUDS_STATOUT]
 |             OUT->send status
 |             IN->stall
#endif /* 0 */

/*
 *
 */
#define EP_CTRL_IN   (0x80 | 0)
#define EP_CTRL_OUT  (0x00 | 0)

/*
 * functions
 */
int ucore_InitialCore(void);
void ucore_ResetCore(void);
void ucore_WakeUp_host(void);
void ucore_feed_SetupCommand(void);
void ucore_SetupCommand(void);
static void ucore_start_DataIn(void);
static void ucore_start_DataOut(void);
static void ucore_start_StatusIn(void);
static int ucore_GetStatus(void);
static int ucore_SetClrFeature(u_long sc);
static int ucore_GetDescriptor(void);
static u_char *ucore_GetStringDscr(u_char* top_string_desc, u_char index);
static int ucore_SetConfiguration(void);
static int ucore_SetInterface(void);
void ucore_isr_Ep0(u_long event);
static void ucore_DataInStage(void);
static int ucore_DataOutStage(void);
static void ucore_StatusInStage(void);
static void ucore_StatusOutStage(void);
void ucore_STALL_EP0(void);

/*
 * variables
 */
u_char *USB_DeviceDescriptor;
u_char *USB_ConfigDescriptor;
u_char *USB_StringDescriptor;
u_long max_packet_size_ep0;
           u_short USB_DeviceStatus;
/*static*/ u_char  USB_DeviceAddress;
           u_char  USB_Configuration;

/*static*/ u_long  USB_EndPointMask;
/*static*/ u_long  USB_EndPointHalt;
/*static*/ u_char  USB_NumInterfaces;
/*static*/ u_char  USB_AltSetting[USB_N_IF];

/*static*/ int control_pipe_state;
/*static*/ int cpipe_add_zero_packet;
/*static*/ int cpipe_out_data_ready;
/*static*/ u_long cpipe_len_requested;
/*static*/ u_char *cpipe_data;
/*static*/ u_long cpipe_count;
/*static*/ u_long cpipe_out_buff_remain;	/* OUT only */

u_char SETUPDAT[8];
int GotSUD;
/*static*/ u_char EP0BUF[EP0_BUFF_SIZE];

/*
 * externs
 */
/* in dscr.S */
extern u_long *DeviceDscr, *ConfigDscr, *StringDscr;
/* in usbhw.c */
extern int uhw_Init(void);
extern void uhw_WakeUp(void);
extern void uhw_SetAddress(u_char adr);
extern void uhw_Configure(int cfg);
extern void uhw_ConfigEP(u_char *pEPD);
extern void uhw_EnableEP(u_char EPNum);
extern void uhw_DisableEP(u_char EPNum);
extern void uhw_ResetEP(u_char EPNum);
extern void uhw_SetStallEP(u_char EPNum);
extern void uhw_ClrStallEP(u_char EPNum);
extern u_long uhw_WriteEP(u_char EPNum, u_char *pData, u_long cnt);
extern void uhw_arm_WriteEP(u_char EPNum);
extern int uhw_ReadEP(u_char EPNum, u_char *pData, u_long DataSize);
extern void uhw_ReadEP_finish(u_char EPNum);
/* in vcomuser.c */
extern void vcom_USB_Configure_Event(void);
extern void vcom_USB_Interface_Event(void);
extern void vcom_USB_Feature_Event(void);
extern int vcom_SetSIOSetup(u_char Cmd, u_char Data);


/*
 * initial valiables about USB core
 */
int ucore_InitialCore(void)
{
	USB_DeviceDescriptor = (u_char *)&DeviceDscr;
	USB_ConfigDescriptor = (u_char *)&ConfigDscr;
	USB_StringDescriptor = (u_char *)&StringDscr;

	max_packet_size_ep0 = USB_DeviceDescriptor[7];

	control_pipe_state = CPS_IDLE;

	/*
	 * initial usbhw.c (max_packet_size_ep0 needed)
	 */
	return(uhw_Init());
}

/*
 * Reset USB Core
 */
void ucore_ResetCore(void)
{
	USB_DeviceStatus  = USB_POWER;
	USB_DeviceAddress = 0;
	USB_Configuration = 0;

#define EP_MASK_IN(x)  ((1<<16)<<(x))		/* !! shall pass (n & 0x0f) !! */
#define EP_MASK_OUT(x)  (1<<(x))
#define DEF_ENDPOINT_MASK  (EP_MASK_IN(0) | EP_MASK_OUT(0))
	USB_EndPointMask = DEF_ENDPOINT_MASK;	/*  */
	USB_EndPointHalt = 0x00000000;			/*  */
}

/*
 * USB Remote Wakeup Function
 */
void ucore_WakeUp_host(void)
{
	if(USB_DeviceStatus & USB_GETSTATUS_REMOTE_WAKEUP) {
		uhw_WakeUp();
	}
}

/*
 * take SETUP data out from hardware buffer
 */
void ucore_feed_SetupCommand(void)
{
	uhw_ReadEP(EP_CTRL_OUT, SETUPDAT, sizeof(SETUPDAT));
//	uhw_ReadEP_finish(EP_CTRL_OUT);
// try, how act does he ? ==> status OUT is pend !
	GotSUD = FALSE;

//	SetupPacket.dir   = (SETUPDAT[0] >> 7) & 0x01;
//	SetupPacket.type  = (SETUPDAT[0] >> 5) & 0x03;
//	SetupPacket.recip = (SETUPDAT[0] >> 0) & 0x1f;
//	SetupPacket.bRequest = SETUPDAT[1];
//	SetupPacket.wValue  = SETUPDAT[2] + (SETUPDAT[3] << 8);
//	SetupPacket.wIndex  = SETUPDAT[4] + (SETUPDAT[5] << 8);
//	SetupPacket.wLength = SETUPDAT[6] + (SETUPDAT[7] << 8);

	cpipe_len_requested = SETUPDAT[6] + (SETUPDAT[7] << 8);
}

/*
 * device request parser
 */
void ucore_SetupCommand(void)
{
	switch((SETUPDAT[0] >> 5) & 0x03 /*type*/) {
	case REQUEST_STANDARD:
			switch(SETUPDAT[1]) {
			case USB_REQUEST_GET_STATUS:
rprintf("GET_STAT\r\n");
					if(!ucore_GetStatus()) {
						goto stall_i;
					}
					ucore_start_DataIn();
					break;

			case USB_REQUEST_CLEAR_FEATURE:
rprintf("CLR_FEAT\r\n");
					if(!ucore_SetClrFeature(0)) {
						goto stall_i;
					}
					ucore_start_StatusIn();
//					vcom_USB_Feature_Event();
					break;

			case USB_REQUEST_SET_FEATURE:
rprintf("SET_FEAT\r\n");
					if(!ucore_SetClrFeature(1)) {
						goto stall_i;
					}
					ucore_start_StatusIn();
//					vcom_USB_Feature_Event();
					break;

			case USB_REQUEST_SET_ADDRESS:
rprintf("SET_ADDR\r\n");
					switch(SETUPDAT[0] & 0x1f /*recip*/) {
					case REQUEST_TO_DEVICE:
							USB_DeviceAddress = 0x80 | SETUPDAT[2];
							ucore_start_StatusIn();
							break;
					default:
							goto stall_i;
					}
					break;

			case USB_REQUEST_GET_DESCRIPTOR:
rprintf("GET_DESC\r\n");
					if(!ucore_GetDescriptor()) {
						goto stall_i;
					}
					ucore_start_DataIn();
					break;

			case USB_REQUEST_SET_DESCRIPTOR:
rprintf("SET_DESC\r\n");
					ucore_STALL_EP0();
					cpipe_count = 0;
					break;

			case USB_REQUEST_GET_CONFIGURATION:
rprintf("GET_CONF\r\n");
					switch(SETUPDAT[0] & 0x1f /*recip*/) {
					case REQUEST_TO_DEVICE:
							EP0BUF[0] = USB_Configuration;
							cpipe_data = EP0BUF;
							cpipe_count = 1;
							ucore_start_DataIn();
							break;
					default:
							goto stall_i;
					}
					break;

			case USB_REQUEST_SET_CONFIGURATION:
rprintf("SET_CONF\r\n");
					switch(SETUPDAT[0] & 0x1f /*recip*/) {
					case REQUEST_TO_DEVICE:
							if(!ucore_SetConfiguration()) {
								goto stall_i;
							}
							ucore_start_StatusIn();
//							vcom_USB_Configure_Event();
							break;
					default:
							goto stall_i;
					}
					break;

			case USB_REQUEST_GET_INTERFACE:
rprintf("GET_INTF\r\n");
					switch(SETUPDAT[0] & 0x1f /*recip*/) {
					case REQUEST_TO_INTERFACE:
							if(
								(USB_Configuration != 0) &&
								(SETUPDAT[4] < USB_NumInterfaces)
							) {
								EP0BUF[0] = USB_AltSetting[SETUPDAT[4]];
								cpipe_data = EP0BUF;
								cpipe_count = 1;
								ucore_start_DataIn();
							} else {
								goto stall_i;
							}
							break;
					default:
							goto stall_i;
					}
					break;

			case USB_REQUEST_SET_INTERFACE:
rprintf("SET_INTF\r\n");
					switch(SETUPDAT[0] & 0x1f /*recip*/) {
					case REQUEST_TO_INTERFACE:
							if(!ucore_SetInterface()) {
								goto stall_i;
							}
							ucore_start_StatusIn();
//							vcom_USB_Interface_Event();
							break;
					default:
							goto stall_i;
					}
					break;

			default:
					goto stall_i;
			}
			break;

	case REQUEST_CLASS:
#if UDR_NEED_CLASS_SUPPORT
			switch(SETUPDAT[0] & 0x1f /*recip*/) {
			case REQUEST_TO_INTERFACE:
					//break;
			default:
					goto stall_i;
			}
#else /* UDR_NEED_CLASS_SUPPORT */
			goto stall_i;
#endif /* UDR_NEED_CLASS_SUPPORT */

	case REQUEST_VENDOR:
#if UDR_NEED_VENDOR_SUPPORT
			UARTSettingCmd = SetupPacket.bRequest & 0xFF;
			UARTSettingData = SetupPacket.wValue & 0xFF;
			if(!vcom_SetSIOSetup( UARTSettingCmd, UARTSettingData )) {
				goto stall_i;
			}
			ucore_start_DataIn();
#endif /* UDR_NEED_VENDOR_SUPPORT */
			break;

	default:
stall_i:
			ucore_STALL_EP0();
			cpipe_count = 0;
			break;
	}
rprintf("@\r\n");
}

/*
 * initiate data IN
 */
static void ucore_start_DataIn(void)
{
	/*
	 * round data length by request length
	 */
	if(cpipe_count > cpipe_len_requested) {
		cpipe_count = cpipe_len_requested;
	}

	/*
	 * I expect IN from the host
	 */
	control_pipe_state = CPS_DATAIN;
	cpipe_add_zero_packet = FALSE;		/* for IN data */

	/*
	 * parse the request finished, but keep OUT buffer filled,
	 * no OUT packet should come until status-out phase
	 */
//	uhw_ReadEP_finish(EP_CTRL_OUT);

	/*
	 * initiate first data-in packet
	 */
	ucore_DataInStage();
}

/*
 * initiate data OUT
 */
static void ucore_start_DataOut(void)
{
	/*
	 * round data length by request length
	 */
	if(cpipe_count > cpipe_len_requested) {
		cpipe_count = cpipe_len_requested;
	}

	/*
	 * I expect OUT from the host
	 */
	control_pipe_state = CPS_DATAOUT;
	cpipe_out_data_ready = FALSE;		/* for OUT data */

	/*
	 * parse the request finished, make out buffer free
	 */
	uhw_ReadEP_finish(EP_CTRL_OUT);

	/*
	 * wait for OUT packet will come
	 */
}

/*
 * initiate status IN
 */
static void ucore_start_StatusIn(void)
{
	/*
	 * I expect IN from the host
	 */
	control_pipe_state = CPS_STATIN;

	/*
	 * parse the request finished, but keep OUT buffer filled,
	 * no OUT packet should come until status-out phase
	 */
//	uhw_ReadEP_finish(EP_CTRL_OUT);

	/*
	 * initiate status-in packet
	 */
	ucore_StatusInStage();
}

/*
 * Get Status USB Request
 */
static int ucore_GetStatus(void)
{
	u_long n, m;

	switch(SETUPDAT[0] & 0x1f /*recip*/) {
	case REQUEST_TO_DEVICE:
			EP0BUF[0] = (USB_DeviceStatus>>0) & 0xff;
			EP0BUF[1] = (USB_DeviceStatus>>8) & 0xff;
			break;

	case REQUEST_TO_INTERFACE:
			if(
				(USB_Configuration != 0) &&
				(SETUPDAT[4] < USB_NumInterfaces)
			) {
				EP0BUF[0] = 0;
				EP0BUF[1] = 0;
			} else {
				return(FALSE);
			}
			break;

	case REQUEST_TO_ENDPOINT:
			n = SETUPDAT[4] & 0x8F;
			m = (n & 0x80) ? EP_MASK_IN(n) : EP_MASK_OUT(n);
			if(
				((USB_Configuration != 0) || ((n & 0x0F) == 0)) &&
				(USB_EndPointMask & m)
			) {
				EP0BUF[0] = (USB_EndPointHalt & m) ? 1 : 0;
				EP0BUF[1] = 0;
			} else {
				return(FALSE);
			}
			break;

	default:
			return(FALSE);
	}

	cpipe_data = EP0BUF;
	cpipe_count = 2;
	return(TRUE);
}

/*
 * Set/Clear Feature USB Request
 *   pass sc as: 0 for Clear, 1 for Set
 */
static int ucore_SetClrFeature(u_long sc)
{
	u_long n, m;

	switch(SETUPDAT[0] & 0x1f /*recip*/) {
	case REQUEST_TO_DEVICE:
			if((SETUPDAT[2]+(SETUPDAT[3]<<8)) == USB_FEATURE_REMOTE_WAKEUP) {
				if(sc) {
					USB_DeviceStatus |=  USB_GETSTATUS_REMOTE_WAKEUP;
				} else {
					USB_DeviceStatus &= ~USB_GETSTATUS_REMOTE_WAKEUP;
				}
			} else {
				return(FALSE);
			}
			break;

	case REQUEST_TO_INTERFACE:
			return(FALSE);

	case REQUEST_TO_ENDPOINT:
			n = SETUPDAT[4] & 0x8F;
			m = (n & 0x80) ? EP_MASK_IN(n) : EP_MASK_OUT(n);
			if(
				(USB_Configuration != 0) &&
				((n & 0x0F) != 0) &&
				(USB_EndPointMask & m)
			) {
				if((SETUPDAT[2]+(SETUPDAT[3]<<8)) == USB_FEATURE_ENDPOINT_STALL) {
					if(sc) {
						uhw_SetStallEP(n);
						USB_EndPointHalt |=  m;
					} else {
						uhw_ClrStallEP(n);
						USB_EndPointHalt &= ~m;
					}
				} else {
					return(FALSE);
				}
			} else {
				return(FALSE);
			}
			break;

	default:
			return(FALSE);
	}
	return(TRUE);
}

/*
 * Get Descriptor USB Request
 */
static int ucore_GetDescriptor(void)
{
	u_char *pD;
	u_long len, n;
	u_short total_len;

	switch(SETUPDAT[0] & 0x1f /*recip*/) {
	case REQUEST_TO_DEVICE:
			switch(SETUPDAT[3]) {
			case USB_DEVICE_DESCRIPTOR_TYPE:
rprintf(".device\r\n");
					pD = (u_char *)USB_DeviceDescriptor;
					len = pD[0];
					break;

			case USB_CONFIGURATION_DESCRIPTOR_TYPE:
rprintf(".config\r\n");
					pD = (u_char *)USB_ConfigDescriptor;
					for(n=0; n != SETUPDAT[2]; n++) {
						if(pD[/*bLength*/ 0] != 0) {
							total_len = (pD[2]<<0) + (pD[3]<<8);	/* wTotalLength */
							pD += total_len;
						}
					}
					if(pD[/*bLength*/ 0] == 0) {
						return(FALSE);
					}
					len = (pD[2]<<0) + (pD[3]<<8);	/* wTotalLength */
					break;

			case USB_STRING_DESCRIPTOR_TYPE:
rprintf(".string\r\n");
					pD = ucore_GetStringDscr(USB_StringDescriptor, SETUPDAT[2]);
					len = pD[0];		/* bLength */
					break;

			default:
					return(FALSE);
			}
			break;

	case REQUEST_TO_INTERFACE:
			switch(SETUPDAT[3]) {
			default:
					return(FALSE);
			}
			break;

	default:
			return(FALSE);
	}

	cpipe_data = pD;
	cpipe_count = len;
//	if(cpipe_count > len) {
//		cpipe_count = len;
//	}
//
	return(TRUE);
}

/*
 * search Nth string descriptor
 */
static u_char *ucore_GetStringDscr(u_char* top_string_desc, u_char index)
{
	u_char *dscr;

	dscr = top_string_desc;

	while(dscr[1 /*type*/] == USB_STRING_DESCRIPTOR_TYPE) {
		if(index == 0)
			return(dscr);
		index--;
		dscr += dscr[0 /*length*/];
	}

	return(NULL);
}

/*
 * Set Configuration USB Request
 */
static int ucore_SetConfiguration(void)
{
	u_char *pD;
	u_long alt, n, m;
	u_short total_len;

	if(SETUPDAT[2]) {
		pD = (u_char *)USB_ConfigDescriptor;
		while(pD[/*bLength*/ 0]) {
			switch(pD[/*bDescriptorType*/ 1]) {
			case USB_CONFIGURATION_DESCRIPTOR_TYPE:
					if(pD[/*bConfigurationValue*/ 5] == SETUPDAT[2]) {
						USB_Configuration = SETUPDAT[2];
						USB_NumInterfaces = pD[/*bNumInterfaces*/ 4];
						for(n=0; n<USB_N_IF; n++) {
							USB_AltSetting[n] = 0;
						}
						for(n=1; n<16; n++) {
							if(USB_EndPointMask & EP_MASK_OUT(n)) {
								uhw_DisableEP(n);
							}
							if(USB_EndPointMask & EP_MASK_IN(n)) {
								uhw_DisableEP(n | 0x80);
							}
						}
						USB_EndPointMask = DEF_ENDPOINT_MASK;
						USB_EndPointHalt = 0x00000000;
						uhw_Configure(TRUE);

						if(pD[/*bmAttributes*/ 7] & USB_CONFIG_SELF_POWERED) {
							USB_DeviceStatus |=  USB_GETSTATUS_SELF_POWERED;
						} else {
							USB_DeviceStatus &= ~USB_GETSTATUS_SELF_POWERED;
						}
					} else {
						total_len = (pD[2]<<0) + (pD[3]<<8);	/* wTotalLength */
						pD += total_len;
						continue;
					}
					break;

			case USB_INTERFACE_DESCRIPTOR_TYPE:
					alt = pD[/*bAlternateSetting*/ 3];
					break;

			case USB_ENDPOINT_DESCRIPTOR_TYPE:
					if(alt == 0) {
						n = pD[/*bEndpointAddress*/ 2] & 0x8F;
						m = (n & 0x80) ? EP_MASK_IN(n) : EP_MASK_OUT(n);
						USB_EndPointMask |= m;
						uhw_ConfigEP(pD);
						uhw_EnableEP(n);
						uhw_ResetEP(n);
					}
					break;
			}

			pD += pD[/*bLength*/ 0];
		}
	} else {
		USB_Configuration = 0;
		for(n=1; n<16; n++) {
			if(USB_EndPointMask & EP_MASK_OUT(n)) {
				uhw_DisableEP(n);
			}
			if(USB_EndPointMask & EP_MASK_IN(n)) {
				uhw_DisableEP(n | 0x80);
			}
		}
		USB_EndPointMask  = DEF_ENDPOINT_MASK;
		USB_EndPointHalt  = 0x00000000;
		uhw_Configure(FALSE);
	}

	if(USB_Configuration == SETUPDAT[2]) {
		return(TRUE);
	} else {
		return(FALSE);
	}
}

/*
 * Set Interface USB Request
 */
static int ucore_SetInterface(void)
{
	u_char *pD;
	u_long ifn, alt, old, msk, n, m;
	int set;
	u_short total_len;

	if(USB_Configuration == 0)
		return(FALSE);

	set = FALSE;
	pD = (u_char *)USB_ConfigDescriptor;

	while(pD[/*bLength*/ 0]) {
		switch(pD[/*bDescriptorType*/ 1]) {
		case USB_CONFIGURATION_DESCRIPTOR_TYPE:
				if(pD[/*bConfigurationValue*/ 5] != USB_Configuration) {
					total_len = (pD[2]<<0) + (pD[3]<<8);	/* wTotalLength */
					pD += total_len;
					continue;
				}
				break;

		case USB_INTERFACE_DESCRIPTOR_TYPE:
				ifn = pD[/*bInterfaceNumber*/ 2];
				alt = pD[/*bAlternateSetting*/ 3];
				msk = 0;
				if(
					(ifn == SETUPDAT[4]) &&
					(alt == SETUPDAT[2])
				) {
					set = TRUE;
					old = USB_AltSetting[ifn];
					USB_AltSetting[ifn] = (u_char)alt;
				}
				break;

		case USB_ENDPOINT_DESCRIPTOR_TYPE:
				if(ifn == SETUPDAT[4]) {
					n = pD[/*bEndpointAddress*/ 2] & 0x8F;
					m = (n & 0x80) ? EP_MASK_IN(n) : EP_MASK_OUT(n);
					if(alt == SETUPDAT[2]) {
						USB_EndPointMask |=  m;
						USB_EndPointHalt &= ~m;
						uhw_ConfigEP(pD);
						uhw_EnableEP(n);
						uhw_ResetEP(n);
						msk |= m;
					} else if((alt == old) && ((msk & m) == 0)) {
						USB_EndPointMask &= ~m;
						USB_EndPointHalt &= ~m;
						uhw_DisableEP(n);
					}
				}
				break;
		}

		pD += pD[/*bLength*/ 0];
	}
	return(set);
}

/* ==================================================================== */

/*
 * USB endpoint 0 interrupt handler
 */
void ucore_isr_Ep0(u_long event)
{
	switch(event) {
	case USB_EVT_SETUP:
rprintf("EVT_SETUP\r\n");
			GotSUD = TRUE;
			break;

	case USB_EVT_OUT:
rprintf("EVT_OUT\r\n");
			if(control_pipe_state == CPS_DATAOUT) {
				if(cpipe_count != 0) {
					if(ucore_DataOutStage() == 0) {
						if(cpipe_count == 0) {
							/*
							 * make status-in ready
							 */
//							control_pipe_state = CPS_STATIN;
//							ucore_StatusInStage();
							control_pipe_state = CPS_IDLE;
							cpipe_out_data_ready = TRUE;
						} else {
							/*
							 * permit to receive next OUT packet
							 */
							uhw_ReadEP_finish(EP_CTRL_OUT);
						}
					} else {
						/* ignore this packet */
						uhw_ReadEP_finish(EP_CTRL_OUT);
						control_pipe_state = CPS_IDLE;
						ucore_STALL_EP0();
					}
				} else {
					/* ignore this packet */
					uhw_ReadEP_finish(EP_CTRL_OUT);
					control_pipe_state = CPS_IDLE;
					ucore_STALL_EP0();
				}
			}
			else if(control_pipe_state == CPS_STATOUT) {
				/*
				 * always ignore 'status-OUT-has-come'
				 *   (following next SETUP shall be always
				 *    'Packet over-written', don't mind)
				 */
				//ucore_StatusOutStage();
				//control_pipe_state = CPS_IDLE;
			}
			else {
				control_pipe_state = CPS_IDLE;
				ucore_STALL_EP0();
			}
			break;

	case USB_EVT_IN:
rprintf("EVT_IN\r\n");
			if(control_pipe_state == CPS_DATAIN) {
				if((cpipe_count != 0) || cpipe_add_zero_packet) {
					ucore_DataInStage();
				} else {
					control_pipe_state = CPS_STATOUT;
					/*
					 * permit to receive OUT packet
					 */
					uhw_ReadEP_finish(EP_CTRL_OUT);
				}
			}
			else if(control_pipe_state == CPS_STATIN) {
				/*
				 * 'status-in-finished' here, nothing to do.
				 * only about SET_ADDR request, update device address here.
				 */
				if(USB_DeviceAddress & 0x80) {
					USB_DeviceAddress &= 0x7F;
					uhw_SetAddress(USB_DeviceAddress);
				}
				control_pipe_state = CPS_IDLE;
			}
			else {
				control_pipe_state = CPS_IDLE;
				ucore_STALL_EP0();
			}
			break;
	}
}

/*
 * USB endpoint 0 interrupt handler for data IN 
 */
static void ucore_DataInStage(void)
{
	u_long cnt;

	/*
	 * calc data length for this packet
	 */
	if(cpipe_count > max_packet_size_ep0) {
		cnt = max_packet_size_ep0;
	} else {
		cnt = cpipe_count;
	}
rprintf("DataIn(): [n=%d] %02x %02x..\r\n", cnt, cpipe_data[0], cpipe_data[1]);

	/*
	 * write the data into the buffer
	 */
	uhw_WriteEP(EP_CTRL_IN, cpipe_data, cnt);

	/*
	 *
	 */
	if(cpipe_add_zero_packet) {
		/*
		 * already do this operation, reset this flag
		 */
		cpipe_add_zero_packet = FALSE;
	} else {
		/*
		 * advance pointers for next operation
		 */
		cpipe_data += cnt;
		cpipe_count -= cnt;
		cpipe_len_requested -= cnt;

		/*
		 * add zero-length-packet if the data length is just fit
		 * to packet size while the host expect more.
		 */
		if(
			(cnt == max_packet_size_ep0) &&
			(cpipe_count == 0) &&
			(cpipe_len_requested != 0)
		) {
			cpipe_add_zero_packet = TRUE;
		}
	}

	/*
	 * at last, arm the endpoint
	 *   (to prevent 'transfer-finished' interrupt among above process)
	 */
	uhw_arm_WriteEP(EP_CTRL_IN);
}

/*
 * USB endpoint 0 interrupt handler for data OUT
 */
static int ucore_DataOutStage(void)
{
	int cnt;

	/*
	 * take received data out
	 */
	cnt = uhw_ReadEP(EP_CTRL_OUT, cpipe_data, cpipe_out_buff_remain);
rprintf("DataOut(): [n=%d] %02x %02x..\r\n", cnt, cpipe_data[0], cpipe_data[1]);

	if(cnt < 0) {
		return(-1);		/* buffer overflow */
	}

	/*
	 * advance pointers for next operation
	 */
	cpipe_data += cnt;
	cpipe_count -= cnt;
//	cpipe_len_requested -= cnt;
	cpipe_out_buff_remain -= cnt;

	/*
	 * at last, free the endpoint
	 *   (to prevent next 'data-received' interrupt among above process)
	 */
//	uhw_ReadEP_finish(EP_CTRL_OUT);

	return(0);
}

/*
 * USB endpoint 0 interrupt handler for status IN
 */
static void ucore_StatusInStage(void)
{
rprintf("StatusInStage()\r\n");
	uhw_WriteEP(EP_CTRL_IN, NULL, 0);
	uhw_arm_WriteEP(EP_CTRL_IN);
}

/*
 * USB endpoint 0 interrupt handler for status OUT
 */
static void ucore_StatusOutStage(void)
{
rprintf("StatusOutStage()\r\n");
	uhw_ReadEP(EP_CTRL_OUT, EP0BUF, 0);
	uhw_ReadEP_finish(EP_CTRL_OUT);
}

/*
 * stall both End Point 0 in/out
 */
void ucore_STALL_EP0(void)
{
	uhw_SetStallEP(EP_CTRL_IN);		/* for IN */
	uhw_SetStallEP(EP_CTRL_OUT);	/* for OUT */
rprintf("STALL_EP0()\r\n");
		/*
		 * the stall bits are automatically cleared when next SETUP received
		 */
}
